"""HipChat enabled python logger."""
import logging

from .notifications import notify_room


class HipChatHandler(logging.Handler):

    """Log handler used to send notifications to HipChat."""

    DEFAULT_COLOURS = {
        'DEBUG': 'gray',
        'INFO': 'yellow',
        'WARN': 'purple',
        'WARNING': 'purple',
        'ERROR': 'red',
        'CRITICAL': 'red',
    }

    def __init__(
        self,
        token,
        room,
        label='',
        notify=False,
        colors=DEFAULT_COLOURS,
        message_format='html'
    ):
        """
        Args:
            token: the auth token for access to the API - see hipchat.com
            room: the numerical id of the room to send the message to

        Kwargs:
            sender : the 'from' property of the message - appears in the HipChat window
            notify : if True, HipChat pings / bleeps / flashes when message is received
            colors : a dict of level:color pairs (e.g. {'DEBUG:'red'} used to
                override the default color)

        """
        logging.Handler.__init__(self)
        self.token = token
        self.room = room
        self.label = label
        self.notify = notify
        self.colors = colors
        self.message_format = message_format

    def emit(self, record):
        """Send the record info to HipChat."""
        notify_room(
            self.room,
            record.getMessage(),
            color=self.colors.get(record.levelname, 'yellow'),
            label=self.label,
            notify=self.notify,
            message_format=self.message_format
        )



# -*- coding: utf-8 -*-
from dateutil import parser
import decimal
import json
import os


class RequiredSettingMissing(Exception):

    def __init__(self, key):
        msg = u"Required env var '%s' is missing." % key
        super(RequiredSettingMissing, self).__init__(msg)


def _get_env(key, default=None, coerce=lambda x: x, required=False):
    """Return env var coerced into a type other than string.

    This function extends the standard os.getenv function to enable
    the coercion of values into data types other than string (all env
    vars are strings be default).

    Args:
        key: string, the name of the env var to look up

    Kwargs:
        default: the default value to return if the env var does not exist
        coerce: a function that is used to coerce the value returned into
            another type
        required: bool, if True, then a RequiredSettingMissing error is raised
            if the env var does not exist.

    Returns the env var, passed through the coerce function

    """
    try:
        return coerce(os.environ[key])
    except KeyError:
        if required is True:
            raise RequiredSettingMissing(key)
        else:
            return default


# standard type coercian functions
coerce_bool = lambda x: x is not None and x.lower() in ("true", "1", "y")
coerce_int = lambda x: int(x)
coerce_float = lambda x: float(x)
coerce_decimal = lambda x: decimal.Decimal(x)
coerce_dict = lambda x: json.loads(x)
coerce_datetime = lambda x: parser.parse(x)
coerce_date = lambda x: parser.parse(x).date()


def get_env(key, *default, **kwargs):
    """
    Return env var.

    This is the parent function of all other get_foo functions,
    and is responsible for unpacking args/kwargs into the values
    that _get_env expects (it is the root function that actually
    interacts with environ).

    Args:
        key: string, the env var name to look up.
        default: (optional) the value to use if the env var does not
            exist. If this value is not supplied, then the env var is
            considered to be required, and a RequiredSettingMissing
            error will be raised if it does not exist.

    Kwargs:
        coerce: a func that may be supplied to coerce the value into
            something else. This is used by the default get_foo functions
            to cast strings to builtin types, but could be a function that
            returns a custom class.

    Returns the env var, coerced if required, and a default if supplied.

    """
    assert len(default) in (0, 1), "Too many args supplied."
    func = kwargs.get('coerce', lambda x: x)
    required = (len(default) == 0)
    default = default[0] if not required else None
    return _get_env(key, default=default, coerce=func, required=required)


def get_bool(key, *default):
    """Return env var cast as boolean."""
    return get_env(key, *default, coerce=coerce_bool)


def get_int(key, *default):
    """Return env var cast as integer."""
    return get_env(key, *default, coerce=coerce_int)


def get_float(key, *default):
    """Return env var cast as float."""
    return get_env(key, *default, coerce=coerce_float)


def get_decimal(key, *default):
    """Return env var cast as Decimal."""
    return get_env(key, *default, coerce=coerce_decimal)


def get_list(key, *default, **kwargs):
    """Return env var as a list."""
    separator = kwargs.get('separator', ' ')
    func = lambda x: x.split(separator)
    return get_env(key, *default, coerce=func)


def get_dict(key, *default):
    """Return env var as a dict."""
    return get_env(key, *default, coerce=coerce_dict)


def get_date(key, *default):
    """Return env var as a date."""
    return get_env(key, *default, coerce=coerce_date)


def get_datetime(key, *default):
    """Return env var as a datetime."""
    return get_env(key, *default, coerce=coerce_datetime)



"""Custom hipchat exception classes."""
# see https://developer.atlassian.com/hipchat/guide/hipchat-rest-api/api-response-codes
BAD_RESPONSE_CODES = (400, 401, 403, 404, 405, 429, 500, 503)


class HipChatError(Exception):

    """Custom error raised when communicating with HipChat API."""

    def __init__(self, response):
        assert response.status_code in BAD_RESPONSE_CODES, (
            u"Invalid HipChatError response.status_code:{}".format(response)
        )
        assert 'error' in response.json(), (
            u"Invalid HipChatError response.json(): {}".format(response)
        )
        self.status_code = response.status_code
        # NB this is brittle and depends on the API error response existing
        # in the correct format. This is by design - if the response format
        # changes we need to know.
        super(HipChatError, self).__init__(response.json()['error']['message'])



# -*- coding: utf-8 -*-
from exceptions import HipChatError
from logger import HipChatHandler
from notifications import (
    gray,
    green,
    grey,
    purple,
    red,
    yellow,
    notify_room,
    notify_user
)



# -*- coding: utf-8 -*-
"""
Wrapper around the HipChat v2 API for sending user / room notifications.

This module contains a couple of generic functions for sending messages
to rooms and users within HipChat.

    >>> hipchat.notify_room('Customer Service', 'This is a message')
    >>> hipchat.notify_user('hugo', 'Hello, Hugo')

It includes some colour-specific helpers for room notifications:

    >>> hipchat.green('Customer service', 'This is a green message')

Requires HIPCHAT_API_TOKEN to be set.

"""
import logging
import os
import random
import requests

from .exceptions import HipChatError

API_V2_ROOT = 'https://api.hipchat.com/v2/'
SEND_USER_MESSAGE_URL = lambda user: "{}user/{}/message".format(API_V2_ROOT, user)
SEND_ROOM_MESSAGE_URL = lambda room: "{}room/{}/notification".format(API_V2_ROOT, room)
VALID_COLORS = ('yellow', 'green', 'red', 'purple', 'gray', 'random')
VALID_FORMATS = ('text', 'html')

logger = logging.getLogger('hipchat')


def _token():
    """
    Get a valid 'personal' auth token from HIPCHAT_API_TOKEN env var.

    The HipChat API is rate limited, and the advice from the support team
    is that we either build a proper HipChat Connect app (which will take
    time, although may the way ahead), OR we use a set of tokens, each of
    which is a valid user token. To this end, the HIPCHAT_API_TOKEN var read
    in from the environ _could_ be a list of comma-separated tokens, in which
    case this function will pick one at random from the list.

    Return a token if one exists or None.

    """
    try:
        token = os.getenv('HIPCHAT_API_TOKEN')
        return random.choice(token.split(',')).strip()
    except AttributeError:
        return None


def _headers(auth_token):
    """
    Return authentication headers for API requests.

    Args:
        auth_token: string, a valid v2 API token.

    Returns a dict that can be passed into requests.post as the
    'headers' dict.

    """
    return {
        'Authorization': 'Bearer {}'.format(auth_token),
        'Host': 'api.hipchat.com',
        'Content-Type': 'application/json'
    }


def _api(
    url,
    message,
    color='yellow',
    label=None,
    notify=False,
    message_format='html'
):
    """
    Send message to user or room via API.

    Args:
        url: API endpoint
        message: The message body (1-10000 chars)

    Kwargs:
        color: background color for message (VALID_COLOR, default='yellow')
        label: label to be shown in addition to the sender's name, (0-64 chars)
        notify: bool (default=False), whether this message should trigger a
            user notification (change the tab color, play a sound, notify
            mobile phones, etc). Each recipient's notification preferences
            are taken into account.
        message_format: determines how the message is rendered inside
            HipChat applications (VALID_FORMAT, default='html')

    Returns HTTP Response object if successful, else raises HipChatError.

    """
    # asserts - message, colour and format are acceptable
    assert message is not None, u"Missing message param"
    assert len(message) >= 1, u"Message too short, must be 1-10,000 chars."
    assert color in VALID_COLORS, u"Invalid color value: {}".format(color)
    assert message_format in VALID_FORMATS, u"Invalid format: {}".format(message_format)

    token = _token()
    if token is None:
        logger.debug("HipChat API token not found, logging message instead:")
        logger.debug(message)
        return
    headers = _headers(auth_token=token)
    data = {
        'message': message[:10000],
        'color': color,
        'notify': notify,
        'message_format': message_format,
        'from': label[:64] if label else ''
    }

    try:
        resp = requests.post(url, json=data, headers=headers)
        resp.raise_for_status()
        return resp
    except requests.HTTPError as ex:
        raise HipChatError(ex.response)


# see https://www.hipchat.com/docs/apiv2/method/send_room_notification
def notify_room(
    room,
    message,
    color='yellow',
    label=None,
    notify=False,
    message_format='html'
):
    """
    Send a room notification via 'Send room notification' API.

    Args:
        room: The id or url encoded name of the room
        message: The message body (1-10,000 chars)

    Kwargs:
        color: background color for message (VALID_COLOR, default='yellow')
        label: label to be shown in addition to the sender's name, (0-64 chars)
        notify: bool (default=False), whether this message should trigger a
            user notification (change the tab color, play a sound, notify
            mobile phones, etc). Each recipient's notification preferences
            are taken into account.
        message_format: determines how the message is rendered inside
            HipChat applications (VALID_FORMAT, default='html')

    Returns HTTP Response object if successful, else raises HipChatError.

    """
    return _api(
        SEND_ROOM_MESSAGE_URL(room),
        message,
        color=color,
        label=label,
        notify=notify,
        message_format=message_format
    )


# see https://www.hipchat.com/docs/apiv2/method/private_message_user
def notify_user(
    user,
    message,
    notify=False,
    message_format='html'
):
    """
    Send a user notification via 'Send private message' API.

    Args:
        user: The id, email address, or mention name of the recipient
        message: the message body (1-10,000 chars)

    Kwargs:
        notify: bool (default=False), whether this message should trigger a
            user notification (change the tab color, play a sound, notify
            mobile phones, etc). Each recipient's notification preferences
            are taken into account.
        message_format: determines how the message is rendered inside
            HipChat applications (VALID_FORMAT, default='html')

    Returns HTTP Response object if successful, else raises HipChatError.

    """
    return _api(
        SEND_USER_MESSAGE_URL(user),
        message,
        notify=notify,
        message_format=message_format
    )


# ================================
# Colour specific helper functions - NB these can only be sent to rooms.
# ================================

def yellow(room, message, **kwargs):
    """Send a yellow message to a room."""
    kwargs['color'] = 'yellow'
    return notify_room(room, message, **kwargs)


def gray(room, message, **kwargs):
    """Send a gray message to a room."""
    kwargs['color'] = 'gray'
    return notify_room(room, message, **kwargs)


# Aliased for UK spelling.
grey = gray


def green(room, message, **kwargs):
    """Send a green message to a room."""
    kwargs['color'] = 'green'
    return notify_room(room, message, **kwargs)


def purple(room, message, **kwargs):
    """Send a purple message to a room."""
    kwargs['color'] = 'purple'
    return notify_room(room, message, **kwargs)


def red(room, message, **kwargs):
    """Send a red message to a room."""
    kwargs['color'] = 'red'
    return notify_room(room, message, **kwargs)
